@@ -5,7 +5,7 @@ import re
 
 import pykka
 
-from mopidy.mpd import exceptions, protocol
+from mopidy.mpd import exceptions, protocol, tokenize
 
 logger = logging.getLogger(__name__)
 
@@ -68,7 +68,7 @@ class MpdDispatcher(object):
         else:
             return response
 
-    ### Filter: catch MPD ACK errors
+    # Filter: catch MPD ACK errors
 
     def _catch_mpd_ack_errors_filter(self, request, response, filter_chain):
         try:
@@ -78,7 +78,7 @@ class MpdDispatcher(object):
                 mpd_ack_error.index = self.command_list_index
             return [mpd_ack_error.get_mpd_ack()]
 
-    ### Filter: authenticate
+    # Filter: authenticate
 
     def _authenticate_filter(self, request, response, filter_chain):
         if self.authenticated:
@@ -88,15 +88,13 @@ class MpdDispatcher(object):
             return self._call_next_filter(request, response, filter_chain)
         else:
             command_name = request.split(' ')[0]
-            command_names_not_requiring_auth = [
-                command.name for command in protocol.mpd_commands
-                if not command.auth_required]
-            if command_name in command_names_not_requiring_auth:
+            command = protocol.commands.handlers.get(command_name)
+            if command and not command.auth_required:
                 return self._call_next_filter(request, response, filter_chain)
             else:
                 raise exceptions.MpdPermissionError(command=command_name)
 
-    ### Filter: command list
+    # Filter: command list
 
     def _command_list_filter(self, request, response, filter_chain):
         if self._is_receiving_command_list(request):
@@ -119,7 +117,7 @@ class MpdDispatcher(object):
             self.command_list_index is not None and
             request != 'command_list_end')
 
-    ### Filter: idle
+    # Filter: idle
 
     def _idle_filter(self, request, response, filter_chain):
         if self._is_currently_idle() and not self._noidle.match(request):
@@ -142,7 +140,7 @@ class MpdDispatcher(object):
     def _is_currently_idle(self):
         return bool(self.context.subscriptions)
 
-    ### Filter: add OK
+    # Filter: add OK
 
     def _add_ok_filter(self, request, response, filter_chain):
         response = self._call_next_filter(request, response, filter_chain)
@@ -153,7 +151,7 @@ class MpdDispatcher(object):
     def _has_error(self, response):
         return response and response[-1].startswith('ACK')
 
-    ### Filter: call handler
+    # Filter: call handler
 
     def _call_handler_filter(self, request, response, filter_chain):
         try:
@@ -164,26 +162,14 @@ class MpdDispatcher(object):
             raise exceptions.MpdSystemError(e)
 
     def _call_handler(self, request):
-        (handler, kwargs) = self._find_handler(request)
+        tokens = tokenize.split(request)
         try:
-            return handler(self.context, **kwargs)
+            return protocol.commands.call(tokens, context=self.context)
         except exceptions.MpdAckError as exc:
             if exc.command is None:
-                exc.command = handler.__name__.split('__', 1)[0]
+                exc.command = tokens[0]
             raise
 
-    def _find_handler(self, request):
-        for pattern in protocol.request_handlers:
-            matches = re.match(pattern, request)
-            if matches is not None:
-                return (
-                    protocol.request_handlers[pattern], matches.groupdict())
-        command_name = request.split(' ')[0]
-        if command_name in [command.name for command in protocol.mpd_commands]:
-            raise exceptions.MpdArgError(
-                'incorrect arguments', command=command_name)
-        raise exceptions.MpdUnknownCommand(command=command_name)
-
     def _format_response(self, response):
         formatted_response = []
         for element in self._listify_result(response):
@@ -241,7 +227,8 @@ class MpdContext(object):
     #: The subsytems that we want to be notified about in idle mode.
     subscriptions = None
 
-    _invalid_playlist_chars = re.compile(r'[\n\r/]')
+    _invalid_browse_chars = re.compile(r'[\n\r]')
+    _invalid_playlist_chars = re.compile(r'[/]')
 
     def __init__(self, dispatcher, session=None, config=None, core=None):
         self.dispatcher = dispatcher
@@ -251,66 +238,95 @@ class MpdContext(object):
         self.core = core
         self.events = set()
         self.subscriptions = set()
-        self._playlist_uri_from_name = {}
-        self._playlist_name_from_uri = {}
+        self._uri_from_name = {}
+        self._name_from_uri = {}
         self.refresh_playlists_mapping()
 
-    def create_unique_name(self, playlist_name):
-        stripped_name = self._invalid_playlist_chars.sub(' ', playlist_name)
+    def create_unique_name(self, name, uri):
+        stripped_name = self._invalid_browse_chars.sub(' ', name)
         name = stripped_name
         i = 2
-        while name in self._playlist_uri_from_name:
+        while name in self._uri_from_name:
+            if self._uri_from_name[name] == uri:
+                return name
             name = '%s [%d]' % (stripped_name, i)
             i += 1
         return name
 
+    def insert_name_uri_mapping(self, name, uri):
+        name = self.create_unique_name(name, uri)
+        self._uri_from_name[name] = uri
+        self._name_from_uri[uri] = name
+        return name
+
     def refresh_playlists_mapping(self):
         """
         Maintain map between playlists and unique playlist names to be used by
         MPD
         """
         if self.core is not None:
-            self._playlist_uri_from_name.clear()
-            self._playlist_name_from_uri.clear()
             for playlist in self.core.playlists.playlists.get():
                 if not playlist.name:
                     continue
                 # TODO: add scheme to name perhaps 'foo (spotify)' etc.
-                name = self.create_unique_name(playlist.name)
-                self._playlist_uri_from_name[name] = playlist.uri
-                self._playlist_name_from_uri[playlist.uri] = name
+                name = self._invalid_playlist_chars.sub(' ', playlist.name)
+                self.insert_name_uri_mapping(name, playlist.uri)
 
     def lookup_playlist_from_name(self, name):
         """
         Helper function to retrieve a playlist from its unique MPD name.
         """
-        if not self._playlist_uri_from_name:
+        if not self._uri_from_name:
             self.refresh_playlists_mapping()
-        if name not in self._playlist_uri_from_name:
+        if name not in self._uri_from_name:
             return None
-        uri = self._playlist_uri_from_name[name]
+        uri = self._uri_from_name[name]
         return self.core.playlists.lookup(uri).get()
 
     def lookup_playlist_name_from_uri(self, uri):
         """
         Helper function to retrieve the unique MPD playlist name from its uri.
         """
-        if uri not in self._playlist_name_from_uri:
+        if uri not in self._name_from_uri:
             self.refresh_playlists_mapping()
-        return self._playlist_name_from_uri[uri]
+        return self._name_from_uri[uri]
 
-    # TODO: consider making context.browse(path) which uses this internally.
-    # advantage would be that all browse requests then go through the same code
-    # and we could prebuild/cache path->uri relationships instead of having to
-    # look them up all the time.
-    def directory_path_to_uri(self, path):
-        parts = re.findall(r'[^/]+', path)
-        uri = None
-        for part in parts:
-            for ref in self.core.library.browse(uri).get():
-                if ref.type == ref.DIRECTORY and ref.name == part:
-                    uri = ref.uri
-                    break
-            else:
-                raise exceptions.MpdNoExistError()
-        return uri
+    def browse(self, path, recursive=True, lookup=True):
+        path_parts = re.findall(r'[^/]+', path or '')
+        root_path = '/'.join([''] + path_parts)
+
+        if root_path not in self._uri_from_name:
+            uri = None
+            for part in path_parts:
+                for ref in self.core.library.browse(uri).get():
+                    if (ref.type in (ref.DIRECTORY, ref.ALBUM, ref.PLAYLIST)
+                            and ref.name == part):
+                        uri = ref.uri
+                        break
+                else:
+                    raise exceptions.MpdNoExistError('Not found')
+            root_path = self.insert_name_uri_mapping(root_path, uri)
+
+        else:
+            uri = self._uri_from_name[root_path]
+
+        if recursive:
+            yield (root_path, None)
+
+        path_and_futures = [(root_path, self.core.library.browse(uri))]
+        while path_and_futures:
+            base_path, future = path_and_futures.pop()
+            for ref in future.get():
+                path = '/'.join([base_path, ref.name.replace('/', '')])
+                path = self.insert_name_uri_mapping(path, ref.uri)
+
+                if ref.type in (ref.DIRECTORY, ref.ALBUM, ref.PLAYLIST):
+                    yield (path, None)
+                    if recursive:
+                        path_and_futures.append(
+                            (path, self.core.library.browse(ref.uri)))
+                elif ref.type == ref.TRACK:
+                    if lookup:
+                        yield (path, self.core.library.lookup(ref.uri))
+                    else:
+                        yield (path, ref)
